; Patch code for the 3D View (Engine.exe)

proc NewKeybindingsHandler, .hDINPUT, .cbObjectData, .rgdod, .pdwInOut, .dwFlags
begin

        push    [.dwFlags]
        push    [.pdwInOut]
        push    [.rgdod]
        push    [.cbObjectData]
        push    [.hDINPUT]
        call    DWORD [ecx + 28h]                               ; IDirectInputDevice::GetDeviceData

        test    eax,eax                                         ; bail out if the call failed for whatever reason
        jnz     .fail

        mov     ecx,[.pdwInOut]
        mov     ecx,[ecx]

        jecxz   .nofail                                         ; the method succeeded, but there are no keystrokes to be processed

        mov     edx,[.rgdod]

        pushad

.loop:                                                          ; iterate over all of the keys which need to be processed

        mov     eax,[.cbObjectData]
        lea     ebx,[ecx - 1]
        imul    eax,ebx

        lea     esi,[edx + eax]                                 ; points to the scancode
        mov     al,[esi]                                        ; get the scancode
        mov     bl,al

        cmp     al,11h                                          ; W key
        jnz     $+4
        mov     bl,0C8h                                         ; return up arrow instead

        cmp     al,1Fh                                          ; S key
        jnz     $+4
        mov     bl,0D0h                                         ; return down arrow instead

        cmp     al,1Eh                                          ; A key
        jnz     $+4
        mov     bl,0CBh                                         ; return left arrow instead

        cmp     al,20h                                          ; D key
        jnz     $+4
        mov     bl,0CDh                                         ; return right arrow instead

        cmp     al,2Fh                                          ; V key
        jnz     $+4
        mov     bl,52h                                          ; return numpad 0 instead

        cmp     al,2Ah                                          ; left SHIFT key key
        jnz     $+4
        mov     bl,36h                                          ; return right SHIFT key instead

        cmp     al,1Dh                                          ; left CTRL key
        jnz     $+4
        mov     bl,9Dh                                          ; return right CTRL key instead

        mov     [esi],bl                                        ; put the modified scancode back

        loop    .loop

        popad

.nofail:

        xor     eax,eax

.fail:

        cmp     eax,08007000Ch                                  ; original instruction which was replaced by the detour

        return

endp


VisibleInvisibleTextures:

        pushad

        mov     esi,[esp + 3Ch]                                 ; grab the pointer to the start of pixel data

        test    byte [ebp + 1],2                                ; TEXTURE.FLAGS byte - does this texture have an integrated alpha map?
        jz      .exit

        cmp     byte [ebp + 10h],10h                            ; only 16 bits per pixel textures get processed here
        jnz     .exit

        mov     al,[ebp + 14h]                                  ; number of alpha bits per pixel
        dec     al

        mov     edi,[ebp + 4]                                   ; texture horizontal width
        mov     ecx,[ebp + 8]                                   ; texture vertical height

        imul    ecx,edi                                         ; total number of pixels in texture

        test    al,al                                           ; 1 alpha bit per pixel
        jz      .1abpp

        cmp     al,3                                            ; 4 alpha bits per pixel
        jz      .4abpp

.8abpp:                                                         ; 8 alpha bits per pixel

        mov     dl,0FFh
        jmp     .check

.4abpp:

        mov     dl,0F0h
        jmp     .check

.1abpp:

        mov     dl,80h

.check:

        mov     al,[esi + ecx * 2 - 1]
        and     eax,edx

        test    al,al
        jnz     .exit

        loope   .check                                          ; check ALL of the texture pixels!

        call    [GetTickCount]                                  ; initialize the pseudorandom number generator
        mov     ebx,eax

        mov     ecx,[ebp + 8]                                   ; texture vertical height

        mov     [ebp + 14h],byte 4                              ; override the number of alpha bits per pixel to 4bpp!

.loop_outer:

        push    ecx

        mov     edx,ecx

        dec     edx
        imul    edx,[ebp + 0Ch]                                 ; multiply by texture pitch

        lea     edx,[esi + edx]

        mov     ecx,edi                                         ; texture horizontal width

.loop_inner:

        lea     eax,[edx + ecx * 2 - 2]

        push    ecx

        mov     ecx,ebx                                         ; do some simple pseudorandom stuff
        bswap   ecx
        add     ecx,eax
        rol     ecx,9
        xor     ebx,ecx
        rol     ecx,4
        neg     ecx
        xor     ecx,eax
        xor     ebx,ecx

        and     ch,4Fh                                          ; clamp the upper end of the alpha range
        or      ch,20h                                          ; clamp the lower end of the alpha range

        push    eax

        cmp     [esp + 8],dword 2                               ; outer loop counter
        jbe     .border

        mov     eax,[ebp + 8]
        dec     eax
        cmp     [esp + 8],eax                                   ; outer loop counter
        jae     .border

        cmp     [esp + 4],dword 2                               ; inner loop counter
        jbe     .border

        lea     eax,[edi - 1]
        cmp     [esp + 4],eax                                   ; inner loop counter
        jae     .border

        jmp     .puke

.border:

        mov     cx,0F000h                                       ; black border with 100% opacity

.puke:

        pop     eax

        mov     [eax],cx                                        ; overwrite the texture pixel data with the new values

        pop     ecx

        loop    .loop_inner

        pop     ecx

        loop    .loop_outer

.exit:

        popad

        or      esi,ecx                                         ; 2 instructions which were replaced by the detour
        mov     ecx,[ebp + 4]

        retn